@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-meta/search.c,v 1.146.2.23 2009/08/14 20:54:14 quanah Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-meta/search.c,v 1.146.2.26 2010/04/15 22:22:28 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2009 The OpenLDAP Foundation.
+ * Copyright 1999-2010 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * Portions Copyright 1999-2003 Howard Chu.
  * All rights reserved.
@@ -556,7 +556,7 @@ meta_back_search_start(
 	 * Maps filter
 	 */
 	rc = ldap_back_filter_map_rewrite( dc, op->ors_filter,
-			&mfilter, BACKLDAP_MAP );
+			&mfilter, BACKLDAP_MAP, op->o_tmpmemctx );
 	switch ( rc ) {
 	case LDAP_SUCCESS:
 		break;
@@ -574,7 +574,8 @@ meta_back_search_start(
 	 * Maps required attributes
 	 */
 	rc = ldap_back_map_attrs( &mt->mt_rwmap.rwm_at,
-			op->ors_attrs, BACKLDAP_MAP, &mapped_attrs );
+			op->ors_attrs, BACKLDAP_MAP, &mapped_attrs,
+			op->o_tmpmemctx );
 	if ( rc != LDAP_SUCCESS ) {
 		/*
 		 * this target is no longer candidate
@@ -637,10 +638,10 @@ done:;
 	(void)mi->mi_ldap_extra->controls_free( op, rs, &ctrls );
 
 	if ( mapped_attrs ) {
-		free( mapped_attrs );
+		ber_memfree_x( mapped_attrs, op->o_tmpmemctx );
 	}
 	if ( mfilter.bv_val != op->ors_filterstr.bv_val ) {
-		free( mfilter.bv_val );
+		ber_memfree_x( mfilter.bv_val, op->o_tmpmemctx );
 	}
 	if ( mbase.bv_val != realbase.bv_val ) {
 		free( mbase.bv_val );
@@ -1182,14 +1183,17 @@ really_bad:;
 					for ( cnt = 0; references[ cnt ]; cnt++ )
 						;
 	
-					rs->sr_ref = ch_calloc( sizeof( struct berval ), cnt + 1 );
+					rs->sr_ref = ber_memalloc_x( sizeof( struct berval ) * ( cnt + 1 ),
+						op->o_tmpmemctx );
 	
 					for ( cnt = 0; references[ cnt ]; cnt++ ) {
-						ber_str2bv( references[ cnt ], 0, 1, &rs->sr_ref[ cnt ] );
+						ber_str2bv_x( references[ cnt ], 0, 1, &rs->sr_ref[ cnt ],
+						op->o_tmpmemctx );
 					}
 					BER_BVZERO( &rs->sr_ref[ cnt ] );
 	
-					( void )ldap_back_referral_result_rewrite( &dc, rs->sr_ref );
+					( void )ldap_back_referral_result_rewrite( &dc, rs->sr_ref,
+						op->o_tmpmemctx );
 
 					if ( rs->sr_ref != NULL && !BER_BVISNULL( &rs->sr_ref[ 0 ] ) ) {
 						/* ignore return value by now */
@@ -1198,7 +1202,7 @@ really_bad:;
 						( void )send_search_reference( op, rs );
 						op->o_private = savepriv;
 	
-						ber_bvarray_free( rs->sr_ref );
+						ber_bvarray_free_x( rs->sr_ref, op->o_tmpmemctx );
 						rs->sr_ref = NULL;
 					}
 
@@ -1333,23 +1337,27 @@ really_bad:;
 							for ( cnt = 0; references[ cnt ]; cnt++ )
 								;
 	
-							sr_ref = ch_calloc( sizeof( struct berval ), cnt + 1 );
+							sr_ref = ber_memalloc_x( sizeof( struct berval ) * ( cnt + 1 ),
+								op->o_tmpmemctx );
 	
 							for ( cnt = 0; references[ cnt ]; cnt++ ) {
-								ber_str2bv( references[ cnt ], 0, 1, &sr_ref[ cnt ] );
+								ber_str2bv_x( references[ cnt ], 0, 1, &sr_ref[ cnt ],
+									op->o_tmpmemctx );
 							}
 							BER_BVZERO( &sr_ref[ cnt ] );
 	
-							( void )ldap_back_referral_result_rewrite( &dc, sr_ref );
+							( void )ldap_back_referral_result_rewrite( &dc, sr_ref,
+								op->o_tmpmemctx );
 					
 							if ( rs->sr_v2ref == NULL ) {
 								rs->sr_v2ref = sr_ref;
 
 							} else {
 								for ( cnt = 0; !BER_BVISNULL( &sr_ref[ cnt ] ); cnt++ ) {
-									ber_bvarray_add( &rs->sr_v2ref, &sr_ref[ cnt ] );
+									ber_bvarray_add_x( &rs->sr_v2ref, &sr_ref[ cnt ],
+										op->o_tmpmemctx );
 								}
-								ber_memfree( sr_ref );
+								ber_memfree_x( sr_ref, op->o_tmpmemctx );
 							}
 						}
 
@@ -1706,7 +1714,7 @@ finish:;
 	}
 
 	if ( rs->sr_v2ref ) {
-		ber_bvarray_free( rs->sr_v2ref );
+		ber_bvarray_free_x( rs->sr_v2ref, op->o_tmpmemctx );
 	}
 
 	for ( i = 0; i < mi->mi_ntargets; i++ ) {
@@ -2036,7 +2044,7 @@ remove_oc:;
 				ldap_dnattr_result_rewrite( &dc, attr->a_vals );
 
 			} else if ( attr->a_desc == slap_schema.si_ad_ref ) {
-				ldap_back_referral_result_rewrite( &dc, attr->a_vals );
+				ldap_back_referral_result_rewrite( &dc, attr->a_vals, NULL );
 
 			}
 
@@ -2045,12 +2053,12 @@ remove_oc:;
 				int		rc;
 
 				if ( pretty ) {
-					rc = pretty( attr->a_desc->ad_type->sat_syntax,
+					rc = ordered_value_pretty( attr->a_desc,
 						&attr->a_vals[i], &pval, NULL );
 
 				} else {
-					rc = validate( attr->a_desc->ad_type->sat_syntax,
-						&attr->a_vals[i] );
+					rc = ordered_value_validate( attr->a_desc,
+						&attr->a_vals[i], 0 );
 				}
 
 				if ( rc ) {
@@ -2085,9 +2093,9 @@ remove_oc:;
 			attr->a_nvals = ch_malloc( ( last + 1 ) * sizeof( struct berval ) );
 			for ( i = 0; i<last; i++ ) {
 				/* if normalizer fails, drop this value */
-				if ( attr->a_desc->ad_type->sat_equality->smr_normalize(
+				if ( ordered_value_normalize(
 					SLAP_MR_VALUE_OF_ATTRIBUTE_SYNTAX,
-					attr->a_desc->ad_type->sat_syntax,
+					attr->a_desc,
 					attr->a_desc->ad_type->sat_equality,
 					&attr->a_vals[i], &attr->a_nvals[i],
 					NULL )) {
